# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# References:
# DeiT: https://github.com/facebookresearch/deit
# BEiT: https://github.com/microsoft/unilm/tree/master/beit
# --------------------------------------------------------

import math
import sys
from typing import Iterable, Optional
import torch.nn.functional as F
import torch

from timm.data import Mixup
from timm.utils import accuracy

import util.misc as misc
import util.lr_sched as lr_sched


def train_one_epoch(model: torch.nn.Module, criterion: torch.nn.Module,
                    data_loader_forget: Iterable, data_loader_retain: Iterable, 
                    optimizer: torch.optim.Optimizer,
                    device: torch.device, epoch: int, loss_scaler, max_norm: float = 0,
                    mixup_fn: Optional[Mixup] = None, log_writer=None,
                    args=None):
    model.train(True)
    metric_logger = misc.MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', misc.SmoothedValue(window_size=1, fmt='{value:.6f}'))
    header = 'Epoch: [{}]'.format(epoch)
    print_freq = 20

    accum_iter = args.accum_iter

    optimizer.zero_grad()

    if log_writer is not None:
        print('log_dir: {}'.format(log_writer.log_dir))

    for data_iter_step, (obj_forget, obj_retain) in enumerate(
        metric_logger.log_every(data_loader_forget, data_loader_retain, print_freq, header)):
        
        samples_forget, targets_forget = obj_forget
        samples_retain, targets_retain = obj_retain

        # we use a per iteration (instead of per epoch) lr scheduler
        if data_iter_step % accum_iter == 0:
            lr_sched.adjust_learning_rate(optimizer, data_iter_step / len(data_loader_forget) + epoch, args)

        samples_forget = samples_forget.to(device, non_blocking=True)
        targets_forget = targets_forget.to(device, non_blocking=True)

        samples_retain = samples_retain.to(device, non_blocking=True)
        targets_retain = targets_retain.to(device, non_blocking=True)

        if mixup_fn is not None:
            samples_forget, targets_forget = mixup_fn(samples_forget, targets_forget)
            samples_retain, targets_retain = mixup_fn(samples_retain, targets_retain)
        
        print(samples_forget.shape)
        print(samples_retain.shape)
        
        # 假设 model.forward_features 方法正确处理图像并输出特征序列
        features = model.module.forward_features(samples_forget)

        # 确认 features 形状是否正确，最后一个维度应该是 768
        print(features.shape)  # 应该输出形如 [batch_size, seq_length, 768]

        def forget_closure():
            with torch.cuda.amp.autocast():
                with torch.enable_grad():
                    model.zero_grad()
                    # output = model(samples_forget)
                    # loss = criterion(output, samples_forget)
                    loss, pred, mask = model(samples_forget)
            # loss_value = loss.item()

            if not math.isfinite(loss):
                print("Loss is {}, stopping training".format(loss))
                sys.exit(1)

            loss /= accum_iter
            loss.backward()
            return loss
            
        def retain_closure():
            with torch.cuda.amp.autocast():
                with torch.enable_grad():
                    model.zero_grad()
                    # output = model(samples_retain)
                    # loss = criterion(output, samples_retain)
                    loss, pred, mask = model(samples_retain)
            # loss_value = loss.item()

            if not math.isfinite(loss):
                print("Loss is {}, stopping training".format(loss))
                sys.exit(1)

            loss /= accum_iter
            loss.backward()
            return loss
        

        loss_forget, loss_retain = optimizer.step(
            forget_closure=forget_closure, retain_closure=retain_closure, 
            mode = 'one', g_value=0.1, g_constraint=0.1)
        
        '''
        with torch.cuda.amp.autocast():
            outputs = model(samples)
            loss = criterion(outputs, targets)

        loss /= accum_iter
        loss_scaler(loss, optimizer, clip_grad=max_norm,
                    parameters=model.parameters(), create_graph=False,
                    update_grad=(data_iter_step + 1) % accum_iter == 0)

        '''

        if (data_iter_step + 1) % accum_iter == 0:
            optimizer.zero_grad()

        torch.cuda.synchronize()

        metric_logger.update(loss_forget=loss_forget)
        metric_logger.update(loss_retain=loss_retain)
        min_lr = 10.
        max_lr = 0.
        for group in optimizer.param_groups:
            min_lr = min(min_lr, group["lr"])
            max_lr = max(max_lr, group["lr"])

        metric_logger.update(lr=max_lr)

        loss_forget_reduce = misc.all_reduce_mean(loss_forget)
        loss_retain_reduce = misc.all_reduce_mean(loss_retain)

        if log_writer is not None and (data_iter_step + 1) % accum_iter == 0:
            """ We use epoch_1000x as the x-axis in tensorboard.
            This calibrates different curves when batch size changes.
            """
            epoch_1000x = int((data_iter_step / len(data_loader_forget) + epoch) * 1000)
            log_writer.add_scalar('loss_forget', loss_forget_reduce, epoch_1000x)
            log_writer.add_scalar('loss_retain', loss_retain_reduce, epoch_1000x)
            log_writer.add_scalar('lr', max_lr, epoch_1000x)

    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print("Averaged stats:", metric_logger)
    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}


@torch.no_grad()
def evaluate(data_loader, model, device):
    criterion = torch.nn.CrossEntropyLoss()

    metric_logger = misc.MetricLogger(delimiter="  ")
    header = 'Test:'

    # switch to evaluation mode
    model.eval()

    for batch in metric_logger.log_every(data_loader, 10, header):
        images = batch[0]
        target = batch[-1]
        images = images.to(device, non_blocking=True)
        target = target.to(device, non_blocking=True)

        # compute output
        with torch.cuda.amp.autocast():
            output = model(images)
            loss = criterion(output, target)

        acc1, acc5 = accuracy(output, target, topk=(1, 5))

        batch_size = images.shape[0]
        metric_logger.update(loss=loss.item())
        metric_logger.meters['acc1'].update(acc1.item(), n=batch_size)
        metric_logger.meters['acc5'].update(acc5.item(), n=batch_size)
    # gather the stats from all processes
    metric_logger.synchronize_between_processes()
    print('* Acc@1 {top1.global_avg:.3f} Acc@5 {top5.global_avg:.3f} loss {losses.global_avg:.3f}'
          .format(top1=metric_logger.acc1, top5=metric_logger.acc5, losses=metric_logger.loss))

    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}